@@ -35,11 +35,21 @@ class Entry(caching.Memoizable):
 
     # pylint: disable=too-many-instance-attributes,too-many-public-methods
 
-    def __init__(self, record):
-        """ Construct an Entry wrapper
+    __hash__ = caching.Memoizable.__hash__  # type:ignore
+
+    @staticmethod
+    @utils.stash
+    def load(record: model.Entry):
+        """ Get a pooled Entry wrapper
 
         record -- the index record to use as the basis
         """
+        return Entry(Entry.load.__name__, record)
+
+    def __init__(self, create_key, record):
+        """ Instantiate the Entry wrapper """
+
+        assert create_key == Entry.load.__name__, "Entry must be created with Entry.load()"
 
         LOGGER.debug('init entry %d', record.id)
         self._record = record   # index record
@@ -210,7 +220,7 @@ class Entry(caching.Memoizable):
             for record in query.order_by(model.Entry.local_date,
                                          model.Entry.id):
                 if record.is_authorized(cur_user):
-                    return Entry(record)
+                    return Entry.load(record)
 
                 LOGGER.debug("User unauthorized for entry %d", record.id)
                 tokens.request(cur_user)
@@ -235,7 +245,7 @@ class Entry(caching.Memoizable):
             for record in query.order_by(orm.desc(model.Entry.local_date),
                                          orm.desc(model.Entry.id)):
                 if record.is_authorized(cur_user):
-                    return Entry(record)
+                    return Entry.load(record)
 
                 LOGGER.debug("User unauthorized for entry %d", record.id)
                 tokens.request(cur_user)
@@ -246,7 +256,7 @@ class Entry(caching.Memoizable):
     def category(self):
         """ Get the category this entry belongs to. """
         from .category import Category  # pylint: disable=cyclic-import
-        return Category(self._record.category)
+        return Category.load(self._record.category)
 
     @cached_property
     def title(self) -> typing.Callable[..., str]:
@@ -282,7 +292,7 @@ class Entry(caching.Memoizable):
         try:
             return load_message(filepath)
         except FileNotFoundError:
-            expire_file(filepath)
+            expire_record(self._record)
             empty = email.message.Message()
             empty.set_payload('')
             return empty
@@ -521,6 +531,35 @@ class Entry(caching.Memoizable):
             args,
             search_path=self.search_path)
 
+    @cached_property
+    def attachments(self) -> typing.Callable[..., typing.List]:
+        """ Returns a view of entries that are attached to this one. Takes the
+        standard view arguments. """
+
+        def _get_attachments(order=None, **kwargs) -> typing.List:
+            query = queries.build_query({**kwargs,
+                                         'attachments': self._record
+                                         })
+            if order:
+                query = query.order_by(*queries.ORDER_BY[order])
+            return [Entry.load(e) for e in query]
+
+        return CallableProxy(_get_attachments)
+
+    @cached_property
+    def attached(self) -> typing.Callable[..., typing.List]:
+        """ Get all the entries that have attached this one """
+
+        def _get_attached(order=None, **kwargs) -> typing.List:
+            query = queries.build_query({**kwargs,
+                                         'attached': self._record
+                                         })
+            if order:
+                query = query.order_by(*queries.ORDER_BY[order])
+            return [Entry.load(e) for e in query]
+
+        return CallableProxy(_get_attached)
+
     def _get_footnotes(self, body, more, args) -> str:
         """ get the rendered Markdown footnotes for the entry """
         footnotes: typing.List[str] = []
@@ -636,7 +675,8 @@ def get_entry_id(entry, fullpath, assign_id) -> typing.Optional[int]:
             other_entry = model.Entry.get(id=entry_id)
             if (other_entry
                     and os.path.isfile(other_entry.file_path)
-                    and not os.path.samefile(other_entry.file_path, fullpath)):
+                    and not os.path.samefile(other_entry.file_path, fullpath)
+                    and other_entry.status != model.PublishStatus.DRAFT.value):
                 entry_id = None
             else:
                 other_entry = None
@@ -678,7 +718,7 @@ def get_entry_id(entry, fullpath, assign_id) -> typing.Optional[int]:
     return entry_id
 
 
-def save_file(fullpath: str, entry: email.message.Message):
+def save_file(fullpath: str, entry: email.message.Message, fingerprint: str):
     """ Save a message file out, without mangling the headers """
     from atomicwrites import atomic_write
     with atomic_write(fullpath, overwrite=True) as file:
@@ -691,20 +731,27 @@ def save_file(fullpath: str, entry: email.message.Message):
         print('', file=file)
         file.write(entry.get_payload())
 
+        if utils.file_fingerprint(fullpath) != fingerprint:
+            LOGGER.warning("File %s changed during atomic write; aborting", fullpath)
+            raise RuntimeError("File changed during reindex")
+
+    return True
+
 
 @orm.db_session(retry=5)
-def scan_file(fullpath: str, relpath: typing.Optional[str], assign_id: bool) -> bool:
+def scan_file(fullpath: str, relpath: typing.Optional[str], fixup_pass: int) -> bool:
     """ scan a file and put it into the index
 
     :param fullpath str: The full file path
     :param relpath typing.Optional[str]: The file path relative to the content
         root; if None, this will be inferred
-    :param assign_id bool: Whether to assign an ID and fix up the file
+    :param fixup_pass int: Which iteration of fixing-up we're on
 
     """
     # pylint: disable=too-many-branches,too-many-statements,too-many-locals
 
     try:
+        check_fingerprint = utils.file_fingerprint(fullpath)
         entry = load_message(fullpath)
     except FileNotFoundError:
         # The file doesn't exist, so remove it from the index
@@ -713,7 +760,7 @@ def scan_file(fullpath: str, relpath: typing.Optional[str], assign_id: bool) ->
             expire_record(record)
         return True
 
-    entry_id = get_entry_id(entry, fullpath, assign_id)
+    entry_id = get_entry_id(entry, fullpath, fixup_pass > 0)
     if entry_id is None:
         return False
 
@@ -768,6 +815,9 @@ def scan_file(fullpath: str, relpath: typing.Optional[str], assign_id: bool) ->
     values['local_date'] = entry_date.naive
 
     LOGGER.debug("getting entry %s with id %d", fullpath, entry_id)
+
+    remove_by_path(fullpath, entry_id)
+
     record = model.Entry.get(id=entry_id)
     if record:
         LOGGER.debug("Reusing existing entry %d", record.id)
@@ -838,33 +888,74 @@ def scan_file(fullpath: str, relpath: typing.Optional[str], assign_id: bool) ->
 
         orm.commit()
 
+    result = True
+
+    # manage entry attachments
+    with orm.db_session:
+        from .category import search_path as cat_search_path
+        search_path = (os.path.dirname(fullpath), cat_search_path(record.category))
+
+        set_attach = set()
+        for attach in entry.get_all('Attach', []):
+            other = links.find_entry(attach, search_path)
+            if other:
+                set_attach.add(other)
+            elif fixup_pass < 3:
+                # The entry hasn't been found, so treat this as a fixup task
+                # Pass 0 - this entry might not have an ID
+                # Pass 1 - the other entry might not have an ID (since this can be scheduled
+                #    before pass 1 of the other entry)
+                # Pass 2 - everything should have an ID now
+                LOGGER.info("Attempted to link to unknown entry '%s -> %s'; retrying",
+                            relpath, attach)
+                result = False
+            else:
+                LOGGER.warning(
+                    "Failed to link to unknown entry '%s -> %s'; ignoring", relpath, attach)
+
+        remove_attach = []
+        for attach in record.attachments:
+            if attach not in set_attach:
+                remove_attach.append(attach)
+
+        LOGGER.debug("set_attach %s remove_attach %s", set_attach, remove_attach)
+        for attach in remove_attach:
+            record.attachments.remove(attach)
+        for attach in set_attach:
+            record.attachments.add(attach)
+
+        orm.commit()
+
+    # do final fixups
     if record.status == model.PublishStatus.DRAFT.value:
         LOGGER.info("Not touching draft entry %s", fullpath)
     elif fixup_needed:
         LOGGER.info("Fixing up entry %s", fullpath)
-        save_file(fullpath, entry)
+        result = save_file(fullpath, entry, check_fingerprint)
 
-    return True
+    return result
 
 
-@orm.db_session(retry=5)
-def expire_file(filepath):
-    """ Expire a record for a missing file """
-
-    # SQLite doesn't support cascading deletes so clean up manually
-    orm.delete(pa for pa in model.PathAlias if pa.entry.file_path == filepath)
-
-    orm.delete(item for item in model.Entry if item.file_path == filepath)
-    orm.commit()
-
-
-@orm.db_session(retry=5)
+@orm.db_session
 def expire_record(record):
     """ Expire a record for a missing entry """
 
-    # This entry no longer exists so delete it, and anything that references it.
-    # SQLite doesn't support cascading deletes so let's just clean up manually.
+    # This entry no longer exists so delete anything that relies on it
     orm.delete(pa for pa in model.PathAlias if pa.entry == record)
 
-    record.delete()
+    # mark the entry as GONE to remove it from indexes
+    record.status = model.PublishStatus.GONE.value
+    orm.commit()
+
+
+@orm.db_session
+def remove_by_path(fullpath: str, entry_id: int):
+    """ Remove entries for a path that don't match the expected ID """
+
+    orm.delete(pa for pa in model.PathAlias  # type:ignore
+               if pa.entry.file_path == fullpath
+               and pa.entry.id != entry_id)
+    orm.delete(e for e in model.Entry  # type:ignore
+               if e.file_path == fullpath
+               and e.id != entry_id)
     orm.commit()
