@@ -11,11 +11,12 @@ import werkzeug.exceptions as http_error
 from flask import redirect, request, send_file, url_for
 from pony import orm
 
-from . import (caching, config, image, index, model, path_alias, queries, user,
-               utils, view)
+from . import (__version__, caching, config, image, index, model, path_alias,
+               queries, user, utils, view)
 from .caching import cache
 from .category import Category
-from .entry import Entry, expire_record
+from .entry import Entry
+from .entry import expire_record as entry_expire_record
 from .template import Template, map_template
 
 LOGGER = logging.getLogger(__name__)
@@ -83,8 +84,8 @@ def image_function(template=None,
         # this too
         path += (category.search_path,)
     if template is not None:
-        path += os.path.join(config.content_folder,
-                             os.path.dirname(template.filename)),
+        path += (os.path.join(config.content_folder,
+                              os.path.dirname(template.filename)),)
 
     return lambda filename: image.get_image(filename, path)
 
@@ -95,7 +96,7 @@ def render_publ_template(template: Template, **kwargs) -> typing.Tuple[str, str]
     Returns tuple of (rendered text, etag)
     """
     @cache.memoize(unless=caching.do_not_cache)
-    def do_render(template: Template, **kwargs) -> typing.Tuple[str, str]:
+    def do_render(template: Template, **kwargs) -> typing.Tuple[str, str, typing.Dict]:
         LOGGER.debug("Rendering template %s with args %s and kwargs %s; caching=%s",
                      template, request.args, kwargs, not caching.do_not_cache)
 
@@ -109,27 +110,35 @@ def render_publ_template(template: Template, **kwargs) -> typing.Tuple[str, str]
         }
 
         text = template.render(**args)
-        return text, caching.get_etag(text)
+        return text, caching.get_etag(text), flask.g.stash
 
-    try:
+    @orm.db_session
+    def latest_entry():
         # Cache-busting query based on most recently-visible entry
         cb_query = queries.build_query({})
-        cb_query = cb_query.filter(lambda e: e.status == model.PublishStatus.SCHEDULED.value)
         cb_query = cb_query.order_by(orm.desc(model.Entry.utc_date))
-        latest = Entry(cb_query.first()) if cb_query.first() else None
-        LOGGER.debug("Most recently-scheduled entry: %s", latest)
+        latest = cb_query.first()
+        if latest:
+            LOGGER.debug("Most recently-scheduled entry: %s", latest)
+            return latest.id
+        return None
 
-        return do_render(template,
-                         user=user.get_active(),
-                         _url=request.url,
-                         _index_time=index.last_modified(),
-                         _latest=latest,
-                         **kwargs)
+    try:
+        flask.g.stash = {}
+        text, etag, flask.g.stash = do_render(
+            template,
+            user=user.get_active(),
+            _url=request.url,
+            _index_time=index.last_indexed(),
+            _latest=latest_entry(),
+            _publ_version=__version__.__version__,
+            **kwargs)
+        return text, etag
     except queries.InvalidQueryError as err:
         raise http_error.BadRequest(str(err))
 
 
-@orm.db_session(retry=5)
+@orm.db_session
 def render_error(category, error_message, error_codes,
                  exception=None,
                  headers=None) -> typing.Tuple[str, int, typing.Dict[str, str]]:
@@ -165,13 +174,14 @@ def render_error(category, error_message, error_codes,
     if template:
         return render_publ_template(
             template,
-            category=Category(category),
+            category=Category.load(category),
             error={'code': error_code, 'message': error_message},
             exception=exception)[0], error_code, headers
 
     return '%d %s' % (error_code, error_message), error_code, headers
 
 
+@orm.db_session
 def render_exception(error):
     """ Catch-all renderer for the top-level exception handler """
 
@@ -181,14 +191,15 @@ def render_exception(error):
     # what the actual category is
     category = request.path[1:]
 
-    qsize = index.queue_length()
-    if isinstance(error, http_error.NotFound) and qsize:
+    qsize = index.queue_size()
+    if isinstance(error, http_error.NotFound) and (qsize or index.in_progress()):
         retry = max(5, qsize / 5)
         return render_error(
             category, "Site reindex in progress", 503,
             exception={
                 'type': 'Service Unavailable',
-                'str': "The site's contents are not fully known; please try again later",
+                'str': "The site's contents are not fully known; please try again later (qs="
+                + str(qsize) + ")",
                 'qsize': qsize
             },
             headers={
@@ -225,7 +236,7 @@ def render_exception(error):
     })
 
 
-@orm.db_session(retry=5)
+@orm.db_session
 def render_path_alias(path):
     """ Render a known path-alias (used primarily for forced .php redirects) """
 
@@ -271,7 +282,7 @@ def render_category_path(category: str, template: typing.Optional[str]):
             raise http_error.NotFound("No such category")
 
     if not template:
-        template = Category(category).get('Index-Template') or 'index'
+        template = Category.load(category).get('Index-Template') or 'index'
 
     tmpl = map_template(category, template)
 
@@ -290,11 +301,11 @@ def render_category_path(category: str, template: typing.Optional[str]):
 
     view_spec = view.parse_view_spec(request.args)
     view_spec['category'] = category
-    view_obj = view.View(view_spec)
+    view_obj = view.View.load(view_spec)
 
     rendered, etag = render_publ_template(
         tmpl,
-        category=Category(category),
+        category=Category.load(category),
         view=view_obj)
 
     if request.if_none_match.contains(etag):
@@ -304,6 +315,7 @@ def render_category_path(category: str, template: typing.Optional[str]):
                       'ETag': etag}
 
 
+@orm.db_session
 def render_login_form(redir=None, **kwargs):
     """ Renders the login form using the mapped login template """
 
@@ -333,7 +345,9 @@ def handle_unauthorized(cur_user, category='', **kwargs):
                 "User {name} does not have access".format(name=cur_user.name))
 
         # Render the category's unauthorized template
-        return render_publ_template(tmpl, category=Category(category), **kwargs)[0], 403, NO_CACHE
+        return render_publ_template(tmpl,
+                                    category=Category.load(category),
+                                    **kwargs)[0], 403, NO_CACHE
 
     # User is not already logged in, so present a login page
     raise http_error.Unauthorized()
@@ -349,7 +363,7 @@ def _check_authorization(record, category):
 
         if not authorized:
             return handle_unauthorized(cur_user,
-                                       entry=Entry(record),
+                                       entry=Entry.load(record),
                                        category=category)
     return None
 
@@ -401,7 +415,7 @@ def render_entry(entry_id, slug_text='', category=''):
 
     # see if the file still exists
     if record and not os.path.isfile(record.file_path):
-        expire_record(record)
+        entry_expire_record(record)
         record = None
 
     if not record:
@@ -449,14 +463,14 @@ def render_entry_record(record: model.Entry, category: str, template: typing.Opt
         return redirect(record.redirect_url)
 
     # Get the viewable entry
-    entry_obj = Entry(record)
+    entry_obj = Entry.load(record)
 
     # does the entry-id header mismatch? If so the old one is invalid
     try:
         current_id: typing.Optional[int] = int(entry_obj.get('Entry-ID'))  # type:ignore
     except (KeyError, TypeError, ValueError) as err:
         LOGGER.debug("Error checking entry ID: %s", err)
-        current_id = None
+        current_id = record.id
     if current_id != record.id:
         LOGGER.debug("entry %s says it has id %d, index says %d",
                      entry_obj.file_path, current_id, record.id)
@@ -464,11 +478,11 @@ def render_entry_record(record: model.Entry, category: str, template: typing.Opt
         from .entry import scan_file
         scan_file(entry_obj.file_path, None, True)
 
-        return redirect(url_for('entry', entry_id=record.id))
+        return redirect(url_for('entry', entry_id=current_id))
 
     entry_template = (template
                       or entry_obj.get('Entry-Template')
-                      or Category(category).get('Entry-Template')
+                      or Category.load(category).get('Entry-Template')
                       or 'entry')
 
     tmpl = map_template(category, entry_template)
@@ -478,7 +492,7 @@ def render_entry_record(record: model.Entry, category: str, template: typing.Opt
     rendered, etag = render_publ_template(
         tmpl,
         entry=entry_obj,
-        category=Category(category))
+        category=Category.load(category))
 
     if request.if_none_match.contains(etag):
         return 'Not modified', 304
@@ -535,7 +549,7 @@ def render_transparent_chit():
                        'Last-Modified': 'Tue, 31 Jul 1990 08:00:00 -0000'}
 
 
-@orm.db_session(retry=5)
+@orm.db_session
 def retrieve_asset(filename):
     """ Retrieves a non-image asset associated with an entry """
 
