@@ -6,10 +6,9 @@ import typing
 
 import arrow
 import flask
-from pony import orm
 from werkzeug.utils import cached_property
 
-from . import caching, model, queries, tokens, user, utils
+from . import caching, queries, tokens, user, utils
 from .entry import Entry
 
 ViewSpec = typing.Dict[str, typing.Any]  # pylint:disable=invalid-name
@@ -28,20 +27,6 @@ PAGINATION_PRIORITY = ['date', 'count']
 # All spec keys that indicate a pagination
 PAGINATION_SPECS = OFFSET_PRIORITY + PAGINATION_PRIORITY
 
-#: Ordering queries for different sort orders
-ORDER_BY = {
-    'newest': (orm.desc(model.Entry.local_date), orm.desc(model.Entry.id)),
-    'oldest': (model.Entry.local_date, model.Entry.id),
-    'title': (model.Entry.sort_title, model.Entry.id)
-}
-
-REVERSE_ORDER_BY = {
-    'newest': (model.Entry.local_date, model.Entry.id),
-    'oldest': (orm.desc(model.Entry.local_date), orm.desc(model.Entry.id)),
-    'title': (orm.desc(model.Entry.sort_title), orm.desc(model.Entry.id))
-}
-
-
 SPAN_FORMATS = {
     'day': 'YYYY/MM/DD',
     'week': 'YYYY/MM/DD',
@@ -54,7 +39,11 @@ class View(caching.Memoizable):
     # pylint: disable=too-many-instance-attributes,too-many-public-methods
     """ A view of entries """
 
-    def __init__(self, input_spec: ViewSpec):
+    __hash__ = caching.Memoizable.__hash__  # type:ignore
+
+    @staticmethod
+    @utils.stash
+    def load(input_spec: ViewSpec):
         """ Generate a view.
 
         input_spec -- the parameters to the view. In addition to the values provided
@@ -62,6 +51,12 @@ class View(caching.Memoizable):
             count -- How many entries to include in a page
             order -- How to order the entries ('newest' or 'oldest')
         """
+        return View(View.load.__name__, input_spec)
+
+    def __init__(self, create_key, input_spec: ViewSpec):
+        """ Instantiate the view """
+
+        assert create_key == View.load.__name__, "View must be created with View.load()"
 
         # filter out any priority override things
         spec: ViewSpec = {
@@ -94,7 +89,7 @@ class View(caching.Memoizable):
                 self.spec['last'] = self.spec['start']
 
         self._entries = queries.build_query(
-            spec).order_by(*ORDER_BY[self._order_by])
+            spec).order_by(*queries.ORDER_BY[self._order_by])
 
         if self.spec.get('date') is not None:
             _, self.type, _ = utils.parse_date(self.spec['date'])
@@ -135,7 +130,7 @@ class View(caching.Memoizable):
 
                 auth = record.is_authorized(cur_user)
                 if auth or unauthorized:
-                    result.append(Entry(record))
+                    result.append(Entry.load(record))
                     if not auth and unauthorized is not True:
                         unauthorized -= 1
 
@@ -162,7 +157,7 @@ class View(caching.Memoizable):
 
                 if not record.is_authorized(cur_user):
                     tokens.request(cur_user)
-                    result.append(Entry(record))
+                    result.append(Entry.load(record))
 
             return result
 
@@ -174,7 +169,20 @@ class View(caching.Memoizable):
         query = queries.build_query({**self.spec,
                                      'future': False,
                                      '_deleted': True})
-        return [Entry(e) for e in query]
+        if self.spec.get('count'):
+            # We're using a count constraint, so we'll want to limit this to
+            # entries that come before our next page of non-deleted results.
+            #
+            # We only use the next-page start because our pagination constraint
+            # (if any) will have already limited this query appropriately, and
+            # we consider deleted entries in the gap between two views to
+            # belong to the first page of views.
+            if self._order_by == 'oldest' and self.next:
+                query = queries.where_before_entry(query, self.next.first)
+            elif self._order_by == 'newest' and self.next:
+                query = queries.where_after_entry(query, self.next.first)
+
+        return [Entry.load(e) for e in query]
 
     @cached_property
     def count(self) -> int:
@@ -352,7 +360,7 @@ class View(caching.Memoizable):
         def _get_current(**restrict) -> 'View':
             spec = {k: v for (k, v) in self.spec.items()
                     if k not in OFFSET_PRIORITY}
-            return View({**spec, **restrict})
+            return View.load({**spec, **restrict})
 
         return utils.CallableProxy(_get_current)
 
@@ -371,114 +379,124 @@ class View(caching.Memoizable):
         Returns a tuple of older page, newer page.
         """
 
-        oldest = self.oldest
-        newest = self.newest
-
         base = {key: val for key, val in self.spec.items()
                 if key not in OFFSET_PRIORITY}
 
-        oldest_neighbor = View({
-            **base,
-            'before': oldest,
-            'order': 'newest'
-        }).first if oldest else None
-
-        newest_neighbor = View({
-            **base,
-            'after': newest,
-            'order': 'oldest'
-        }).first if newest else None
-
         if 'date' in self.spec:
-            return self._get_date_pagination(base, oldest_neighbor, newest_neighbor)
+            LOGGER.debug('by date')
+            return self._get_date_pagination(base)
 
         if 'count' in self.spec:
-            return self._get_count_pagination(base, oldest_neighbor, newest_neighbor)
+            LOGGER.debug('by count')
+            return self._get_count_pagination(base)
 
         # we're not paginating
         return None, None
 
-    def _get_date_pagination(self, base: ViewSpec,
-                             oldest_neighbor: typing.Optional[Entry],
-                             newest_neighbor: typing.Optional[Entry]
+    def _get_date_pagination(self,
+                             base: ViewSpec
                              ) -> typing.Tuple[typing.Optional['View'], typing.Optional['View']]:
         """ Compute the pagination for date-based views """
-        _, span, date_format = utils.parse_date(self.spec['date'])
+        date, interval, date_format = utils.parse_date(self.spec['date'])
+        start_date, end_date = date.span(interval)
+
+        base_query = queries.build_query(base)
+        oldest_neighbor = base_query.filter(lambda e: e.local_date < start_date.datetime)\
+            .order_by(*queries.ORDER_BY['newest']).first()
+        newest_neighbor = base_query.filter(lambda e: e.local_date > end_date.datetime)\
+            .order_by(*queries.ORDER_BY['oldest']).first()
 
         older_view: typing.Optional['View'] = None
         newer_view: typing.Optional['View'] = None
 
         if newest_neighbor:
-            newer_date = newest_neighbor.date.span(span)[0]
-            newer_view = View({**base,
-                               'order': self._order_by,
-                               'date': newer_date.format(date_format)})
+            newer_date = newest_neighbor.display_date
+            newer_view = View.load({**base,
+                                    'order': self._order_by,
+                                    'date': arrow.get(newer_date).format(date_format)})
 
         if oldest_neighbor:
-            older_date = oldest_neighbor.date.span(span)[0]
-            older_view = View({**base,
-                               'order': self._order_by,
-                               'date': older_date.format(date_format)})
+            older_date = oldest_neighbor.display_date
+            older_view = View.load({**base,
+                                    'order': self._order_by,
+                                    'date': arrow.get(older_date).format(date_format)})
 
         return older_view, newer_view
 
-    def _get_count_pagination(self, base: ViewSpec,
-                              oldest_neighbor: typing.Optional[Entry],
-                              newest_neighbor: typing.Optional[Entry]
+    def _get_count_pagination(self,
+                              base: ViewSpec
                               ) -> typing.Tuple[typing.Optional['View'], typing.Optional['View']]:
         """ Compute the pagination for count-based views """
 
+        oldest = self.oldest
+        newest = self.newest
+
+        oldest_neighbor = View.load({
+            **base,
+            'before': oldest,
+            'order': 'newest'
+        }).first if oldest else None
+
+        newest_neighbor = View.load({
+            **base,
+            'after': newest,
+            'order': 'oldest'
+        }).first if newest else None
+
         count = self.spec['count']
 
         out_spec = {**base, 'count': count, 'order': self._order_by}
 
         if self._order_by == 'newest':
-            older_view = View({**out_spec,
-                               'last': oldest_neighbor}) if oldest_neighbor else None
+            older_view = View.load({**out_spec,
+                                    'last': oldest_neighbor}) if oldest_neighbor else None
 
-            newer_count = View({**base,
-                                'first': newest_neighbor,
-                                'order': 'oldest',
-                                'count': count}) if newest_neighbor else None
-            newer_view = View({**out_spec,
-                               'last': newer_count.last}) if newer_count else None
+            newer_count = View.load({**base,
+                                     'first': newest_neighbor,
+                                     'order': 'oldest',
+                                     'count': count}) if newest_neighbor else None
+            newer_view = View.load({**out_spec,
+                                    'last': newer_count.last}) if newer_count else None
 
             return older_view, newer_view
 
         if self._order_by == 'oldest':
-            older_count = View({**base,
-                                'last': oldest_neighbor,
-                                'order': 'newest',
-                                'count': count}) if oldest_neighbor else None
-            older_view = View({**out_spec,
-                               'first': older_count.last}) if older_count else None
+            older_count = View.load({**base,
+                                     'last': oldest_neighbor,
+                                     'order': 'newest',
+                                     'count': count}) if oldest_neighbor else None
+            older_view = View.load({**out_spec,
+                                    'first': older_count.last}) if older_count else None
 
-            newer_view = View({**out_spec,
-                               'first': newest_neighbor}) if newest_neighbor else None
+            newer_view = View.load({**out_spec,
+                                    'first': newest_neighbor}) if newest_neighbor else None
 
             return older_view, newer_view
 
         return None, None
 
     def __call__(self, **restrict):
-        return View({**self.spec, **restrict})
+        return View.load({**self.spec, **restrict})
+
+    def __iter__(self):
+        return self.entries().__iter__()
 
     def tag_add(self, *tags: utils.ListLike[str]) -> 'View':
         """ Return a view with the specified tags added """
-        return View({**self.spec, 'tag': self.tags | set(tags)})
+        return View.load({**self.spec, 'tag': self.tags | set(tags)})
 
     def tag_remove(self, *tags: utils.ListLike[str]) -> 'View':
         """ Return a view with the specified tags removed """
-        return View({**self.spec, 'tag': self.tags - set(tags)})
+        return View.load({**self.spec, 'tag': self.tags - set(tags)})
 
     def tag_toggle(self, *tags: utils.ListLike[str]) -> 'View':
         """ Return a view with the specified tags toggled """
-        return View({**self.spec, 'tag': self.tags ^ set(tags)})
+        return View.load({**self.spec, 'tag': self.tags ^ set(tags)})
 
 
 def get_view(**kwargs) -> View:
     """ Wrapper function for constructing a view from scratch """
-    return View(input_spec=kwargs)
+    return View.load(input_spec=kwargs)
 
 
 def parse_view_spec(args) -> ViewSpec:
